package org.steveshipway.blueprints.directive;

import java.io.File;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.logging.log4j.Logger;
import org.steveshipway.blueprints.Util;

import net.minecraft.block.Block;
import net.minecraft.block.state.BlockState;
import net.minecraft.block.state.IBlockState;
import net.minecraft.init.Blocks;
import net.minecraft.item.Item;
import net.minecraft.nbt.JsonToNBT;
import net.minecraft.nbt.NBTBase;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTUtil;
import net.minecraft.server.MinecraftServer;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.BlockPos;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumParticleTypes;
//import net.minecraft.util.ChunkCoordinates;
import net.minecraft.world.World;
//import net.minecraftforge.common.util.ForgeDirection;
import net.minecraft.world.chunk.Chunk;
import net.minecraft.block.properties.IProperty;
import net.minecraft.block.properties.PropertyDirection;

public class BlockDirective extends Directive {
	// this is the main one!
	protected BlockPos offset;
	protected boolean soft = false;
	protected IBlockState block = null;
	protected String nbt = null; // possibly this should be a parsed NBT object
	protected int dv = 0;        // no longer really needed since is included in IBlockState
	protected NBTTagCompound tileNbt;
	
	public BlockDirective(Block b,int dv, boolean soft,BlockPos offset, String nbt) {
		this.offset = offset;
		this.block = b.getBlockState().getBaseState().getBlock().getStateFromMeta(dv);
		this.soft = soft;
		this.nbt = nbt;
		if( nbt != null ) {
			this.tileNbt = Util.parseNBT(nbt,"/",null);
		}
		this.dv = dv;
	}
	public BlockDirective(String definition, String bppath, Logger logger) {
    	Pattern p = Pattern.compile("(?i)^(?:\\s*block)?\\s*(~?)(\\w+:\\w+)(?:\\((\\d+)\\))?\\s+(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(-?\\d+)(?:\\s+(\\{.*\\}|[^\\s]+\\.nbt))?");
    	Matcher m = p.matcher(definition);
    	if( m.find() ) {
    		Block b = null;
    		int dv = 0;
    		if( m.group(1).contains("~") ) {
    			this.soft = true;
    		}

			// coordinates
    		try {
    			this.offset = new  BlockPos(Integer.parseInt( m.group(4) ),Integer.parseInt( m.group(5) ),Integer.parseInt( m.group(6) ));
    		} catch( Exception e ) {
        		logger.error("Cannot parse BLOCK definition: "+definition);
        		return;
    		}

			if( m.group(3) != null ) {
    			dv = Integer.parseInt( m.group(3) );
    		} else { dv = 0; }
    		this.dv = dv;

    		// 2 is the block, 3 is the optional DV
    		// must identify the Block object from the m.group(2) text
			// block specified as module:blockname(dv) {nbt}
    		b = Block.getBlockFromName(m.group(2));
    		if( b == null ) {
    			// invalid name
        		logger.error("Invalid block name in BLOCK def: "+m.group(2));
        		// compatibility
        		b = Block.getBlockFromName("minecraft:stone");
    		}
    		this.block = b.getBlockState().getBaseState();
    		this.block = this.block.getBlock().getStateFromMeta(dv);
			
			this.nbt = m.group(7);
			if( this.nbt != null ) {
				this.tileNbt = Util.parseNBT(this.nbt, bppath, logger);
			}
		
    	} else {
    		logger.error("Cannot parse BLOCK definition: "+definition);
    	}
	}
	
	@Override
	public boolean build(World world, BlockPos location, int r, Logger logger) {
		boolean rv;
		BlockPos blockPos = new BlockPos(location);
		
		if( block == null ) {
    		logger.error("No valid block definition in BLOCK directive");
			return false;
		}
		if( world.isRemote ) {
    		logger.error("Block placement called with remote world?  Should never happen.");
			return false;
		}

		// get block position, rotated
		blockPos = rotateAndAdd(this.offset,r,location);
		
		// add twinkles
		world.spawnParticle(EnumParticleTypes.FIREWORKS_SPARK, (double)blockPos.getX(),(double)blockPos.getY()+1,(double)blockPos.getZ(), 0.0D, 0.2D, 0.0D, null);

		// place the block(s), possibly with the NBT in a tileentity
		//logger.info("Base location  : "+location.posX+","+location.posY+","+location.posZ+" r="+r);
		//logger.info("Offset         : "+this.location.posX+","+this.location.posY+","+this.location.posZ);
		//logger.info("Block placement: "+blockPos.posX+","+blockPos.posY+","+blockPos.posZ);
		
		IBlockState blockstate = world.getBlockState(blockPos);
		IBlockState newstate = this.block;
		newstate = Util.rotateState(newstate,r);
		Block curBlock = blockstate.getBlock();
		int curDv = curBlock.getMetaFromState(blockstate);
		if(this.soft) {
			if( curBlock != Blocks.air && curBlock != Blocks.water && curBlock != Blocks.flowing_water ) {
				return true;  // soft block only placed in air or water
			}
		}
		
		if( curBlock != newstate.getBlock() || curDv != newstate.getBlock().getMetaFromState(newstate) ) {
			rv = world.setBlockState(blockPos,newstate,2);
			if( !rv ) {
				logger.warn("Failed to place block in the world");		
				return false;
			}
		
		}
		
		if( this.tileNbt != null ) {
			NBTTagCompound nbtTag = (NBTTagCompound) this.tileNbt.copy();
			nbtTag.setInteger("x", blockPos.getX());
			nbtTag.setInteger("y", blockPos.getY());
			nbtTag.setInteger("z", blockPos.getZ());
			TileEntity te = TileEntity.createAndLoadEntity(nbtTag);
			world.removeTileEntity(blockPos); // should we instead get and merge?
			world.setTileEntity(blockPos,te);
		}
		
		return true;
	}
	
	@Override
	public String asText() {
		String s = "BLOCK ";
		int thisDv = dv;
		if( this.block == null) {
			return "# BLOCK undefined ";
		}
		if(soft) { 
			s += "~"; 
		}
		s += Block.blockRegistry.getNameForObject(this.block.getBlock());
		// XXXX TODO 
		thisDv = this.block.getBlock().getMetaFromState(this.block);
		if(thisDv>0) {
			s+="("+thisDv+")";
		}
		s += " " + offset.getX() + "," + offset.getY() + "," + offset.getZ();
		if( nbt != null ) {
			s += " " + nbt;
		}
		return s;
	}
}
